@@ -1,19 +1,33 @@
 #!/usr/bin/env python
 # encoding: utf-8
-
+# Copyright 2016-2021 Alexander Mollberg
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+# http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 import collections
 from dataclasses import dataclass
 from enum import Enum
 from pprint import pformat, pprint
-from typing import List, Dict, TypeVar, Generic
+from typing import List, Dict, TypeVar, Generic, Sequence, Tuple
 
 from . import debug
 from .commitdiff import CommitDiff
 from .console_color import *
 from .datastructure_util import flatten, lzip
-from .graph import FileId, update_commit_diff, all_paths
+from .enumerate_paths import all_paths
+from .file_selection import FileSelection
+from .list_dict import StableListDict
 from .spg import Node
-from .stable_list_dict import StableListDict
+from .update import FileId, update_commit_diff
 
 
 # Hierarchy:
@@ -87,6 +101,7 @@ class CellKind(Enum):
   NO_CHANGE = 100
   CHANGE = 101
   BETWEEN_CHANGES = 102
+  BETWEEN_SQUASHABLE = 103
 
 
 @dataclass
@@ -134,8 +149,64 @@ class MultiNodeCell(Cell):
     return not (self == other)
 
 
-def decorate_matrix(m: RowMajorMatrix[Cell]):
-  debug.get('grid').debug("decorate_matrix")
+def changes_at_row(m: RowMajorMatrix[Cell], r: int):
+  n_rows = len(m)
+  if n_rows == 0:
+    return []
+  n_cols = len(m[0])
+  return set([c for c in range(n_cols)
+              if m[r][c].kind == CellKind.CHANGE])
+
+
+def collisions_between(m: RowMajorMatrix[Cell],
+                       row: int,
+                       earlier_row: int):
+  changed_at_end_row = changes_at_row(m, row)
+  for row_i in range(earlier_row + 1, row):
+    if changes_at_row(m, row_i) & changed_at_end_row:
+      return True
+  return False
+
+
+def is_subset_of_earlier(m: RowMajorMatrix[Cell], row: int, earlier_row: int):
+  changed_at_end_row = changes_at_row(m, row)
+  changed_at_earlier_row = changes_at_row(m, earlier_row)
+  return not (changed_at_end_row - changed_at_earlier_row)
+
+
+def find_squashable(m: RowMajorMatrix[Cell]) -> Sequence[Tuple[int, int]]:
+  """
+  Find pairs of squashable commits.
+  Squashable in this case means:
+    * that there are no collisions between the commits, and
+    * that the lines that changes in the later commit are a subset of
+      the lines that changes in the earlier commit.
+  Assumes the matrix has already been decorated with BETWEEN_CHANGES.
+  """
+  n_rows = len(m)
+  for r in range(n_rows):
+    for earlier_r in reversed(range(r)):
+      if collisions_between(m, r, earlier_r):
+        break
+      if is_subset_of_earlier(m, r, earlier_r):
+        yield tuple([earlier_r, r])
+
+
+def mark_squashable(m: RowMajorMatrix[Cell],
+                    squashable_tuples: Sequence[Tuple[int, int]]):
+  """
+  Mark cells between squashable changes.
+  See :py:func: find_squashable
+  """
+  for earlier_r, r in squashable_tuples:
+    changes = changes_at_row(m, r)
+    for r_to_mark in range(earlier_r + 1, r):
+      for c in changes:
+        if m[r_to_mark][c].kind == CellKind.BETWEEN_CHANGES:
+          m[r_to_mark][c].kind = CellKind.BETWEEN_SQUASHABLE
+
+
+def mark_cells_between_changes(m: RowMajorMatrix[Cell]):
   n_rows = len(m)
   if n_rows == 0:
     return m
@@ -159,6 +230,12 @@ def decorate_matrix(m: RowMajorMatrix[Cell]):
         last_patch[c] = r
 
 
+def decorate_matrix(m: RowMajorMatrix[Cell]):
+  debug.get('grid').debug("decorate_matrix")
+  mark_cells_between_changes(m)
+  mark_squashable(m, find_squashable(m))
+
+
 class ConnectionStatus(object):
   EMPTY = 1
   INFILL = 2
@@ -214,7 +291,7 @@ class Fragmap:
   spgs: Dict[FileId, Dict[Node, List[Node]]]
 
   @staticmethod
-  def from_diffs(diffs: List[CommitDiff]):
+  def from_diffs(diffs: List[CommitDiff], files_arg: List[str] = None):
     files = {}
     spgs = {}
     for i, diff in enumerate(diffs):
@@ -224,7 +301,11 @@ class Fragmap:
           debug.get('update').debug(spg.to_dot(file_id))
         debug.get('update').debug("-------")
 
-    return Fragmap(diffs, spgs)
+    selected_files = FileSelection.from_files_arg(files_arg)
+    selected_file_spgs = {file_id: spg
+                          for file_id, spg in spgs.items()
+                          if selected_files.contains(file_id, files)}
+    return Fragmap(diffs, selected_file_spgs)
 
   def patches(self):
     return self._patches
@@ -265,15 +346,16 @@ class Fragmap:
     decorate_matrix(m)
     return m
 
-  def render_for_console(self, colorize):
+  def render_for_console(self, colorize) -> RowMajorMatrix[str]:
     return self._render_for_console(self.generate_matrix(), colorize)
 
-  def _render_for_console(self, matrix: RowMajorMatrix[Cell], colorize: bool):
+  def _render_for_console(self, matrix: RowMajorMatrix[Cell], colorize: bool) \
+          -> RowMajorMatrix[str]:
     n_rows = len(matrix)
     if n_rows == 0:
       return []
     n_cols = len(matrix[0])
-    m = [['.' for _ in range(n_cols)] for _ in range(n_rows)]
+    m = RowMajorMatrix([['.' for _ in range(n_cols)] for _ in range(n_rows)])
 
     def render_cell(cell: SingleNodeCell):
       if cell.kind == CellKind.CHANGE:
@@ -288,6 +370,12 @@ class Fragmap:
           return ANSI_BG_RED + ' ' + ANSI_RESET
         else:
           return '|'
+      if cell.kind == CellKind.BETWEEN_SQUASHABLE:
+        if colorize:
+          # Make background yellow
+          return ANSI_BG_DARK_YELLOW + ' ' + ANSI_RESET
+        else:
+          return '^'
       if cell.kind == CellKind.NO_CHANGE:
         return '.'
       assert False, "Unexpected cell kind: %s" % (cell.kind)
@@ -349,15 +437,15 @@ class BriefFragmap:
       for column_group in column_groups
     ])
 
-  def render_for_console(self, colorize):
+  def render_for_console(self, colorize) -> RowMajorMatrix[str]:
     return self._render_for_console(self.generate_matrix(), colorize)
 
-  def _render_for_console(self, matrix, colorize):
+  def _render_for_console(self, matrix, colorize) -> RowMajorMatrix[str]:
     n_rows = len(matrix)
     if n_rows == 0:
       return []
     n_cols = len(matrix[0])
-    m = [['.' for _ in range(n_cols)] for _ in range(n_rows)]
+    m = RowMajorMatrix([['.' for _ in range(n_cols)] for _ in range(n_rows)])
 
     def render_cell(cell: MultiNodeCell):
       if cell.kind == CellKind.CHANGE:
@@ -371,7 +459,13 @@ class BriefFragmap:
           # Make background red
           return ANSI_BG_RED + ' ' + ANSI_RESET
         else:
-          return '.'
+          return '|'
+      if cell.kind == CellKind.BETWEEN_SQUASHABLE:
+        if colorize:
+          # Make background yellow
+          return ANSI_BG_DARK_YELLOW + ' ' + ANSI_RESET
+        else:
+          return '^'
       if cell.kind == CellKind.NO_CHANGE:
         return '.'
       assert False, "Unexpected cell kind: %s" % (cell.kind)
