@@ -1,9 +1,21 @@
 #!/usr/bin/env python
 # encoding: utf-8
-
+# Copyright 2016-2021 Alexander Mollberg
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+# http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 import unittest
 
-from fragmap.graph import Overlap
+from fragmap.update import Overlap
 from fragmap.span import Span
 
 
